// ARDUINO CODE FOR SMART SATA DRIVER RECEIVER (S.CHABOT)

// SATA OUTER SHAFT IS CONNECTED TO X SOCKET - stepper_outer - driver_outer - address 0b00         (bending & rotating instrument)
// SATA MIDDLE SHAFT IS CONNECTED TO Y SOCKET - stepper_middle - driver_middle - address 0b10      (bending & rotating instrument)
// SATA INNER ROD IS CONNECTED TO Z SOCKET - stepper_inner - driver_inner - address 0b01            (rotating end-effector & rotating instrument)
// TRANSPORTER IS CONNECTED TO E0 SOCKET - stepper_trans - driver_trans - address 0b11              (open-close tip & rotating end-effector & rotating instrument)

// HOMING --> outer and middle shafts are relocated to the exchange position, transporter is relocated to center position
// CALIBRATION --> outer and transporter shafts are rotated to detect physical limits instrument
// INSTRUMENT EXCHANGE --> outer and middle shafts are relocated to the exchange position (difference with homing: transporter is not relocated)

#include <Arduino.h>
#include <HardwareSerial.h>
#include <TMCStepper.h>
#include <SPI.h>
#include <nRF24L01.h>
#include <RF24.h>
#include <AverageValue.h>
#include <AccelStepper.h>

////////////// Wireless communication (radiowaves) ///////////////
RF24 radio(49, 53);  // CE, CSN
const byte address[6] = "00006";

//Create data package to be sent
struct Data_Package {
  bool Up_button;
  bool Right_button;
  bool Down_button;
  bool Left_button;
  bool E_button;
  bool F_button;
  bool Joy_button;
  bool Extra_button;
  bool rotate_CW;
  bool rotate_CCW;
  bool bend_for;
  bool bend_back;
  bool E_stop_button;
  bool Handshake;
};

Data_Package data;  // Create a variable with the above structure

////////////// DEFINE PINS ///////////////
#define X_ENABLE_PIN 38
#define Y_ENABLE_PIN A2
#define Z_ENABLE_PIN A8
#define E0_ENABLE_PIN 24

#define X_DIR_PIN A1   // Direction
#define X_STEP_PIN A0  // Step

#define Y_DIR_PIN A7   // Direction
#define Y_STEP_PIN A6  // Step

#define Z_DIR_PIN 48   // Direction
#define Z_STEP_PIN 46  // Step

#define E0_DIR_PIN 28   // Direction
#define E0_STEP_PIN 26  // Step

#define HALL_MIDDLE_PIN A3  // Hall sensor pin for middle shaft actuator
#define HALL_OUTER_PIN A4   // Hall sensor pin for outer shaft actuator

////////////// VARIABLE FOR STEPPER SETTINGS VIA UART ///////////////
int set_current = 75;
int set_stall = 10;    //Do not set the value too high or the TMC will not detect it. Start low and work your way up
int set_tcools = 800;  //5000 Velocity = 800 TSTEP
int motor_microsteps = 8;
bool run_motor = false;
bool stalled_motor = false;

////////////// VARIABLES TO KEEP TRACK OF BUTTON STATES ///////////////
int startRotate_CW;
int startRotate_CCW;
int startBend_for;
int startBend_back;
int startTip_close;
int startTip_open;
int startRotEnd_CW;
int startRotEnd_CCW;

////////////// BUTTONS THAT INITIALIZE A SEQUENCE ///////////////
int initializeCalibration;
int initializeHoming;
int initializeInstExchange;
int initializeCouple;
int startCalibration;
int startHoming;
int startInstExchange;
int startCouple;

////////////// VARIABLES FOR HALL SENSORS ///////////////
int hall_outer;
int hall_middle;

bool outer_homed = false;
bool middle_homed = false;
bool trans_homed = false;
bool hall_threshold_outer = false;
bool hall_threshold_middle = false;

////////////// VARIABLES TO CHECK COMMUNICATION WITH CONTROLLER IS STILL LIVE ///////////////
int handshake_check;
int prev_handshake_check_state;
int flag;

unsigned long lastHandshakeTime;
unsigned long handshakeInterval = 500;
unsigned long currentTime;

////////////// LOAD MEASURING VALUES ///////////////
const int LoadLimit = 230;          // Adjust the load limit as needed
const int LoadMeasureInterval = 1;  // Load measuring interval in milliseconds
const int LoadMeasureCount = 30;    // Number of load measures to average

uint32_t previous_load_measuring_time = 0;
uint32_t ms = 0;
int load_measures_counter = 0;
int accumulated_load = 0;
int accumulated_load_outer = 0;
int accumulated_load_trans = 0;
int accumulated_load_middle = 0;
int accumulated_load_inner = 0;
int val = 0;
int val_outer = 0;
int val_trans = 0;
int val_middle = 0;
int val_inner = 0;
int setvalflag = 0;
int runonce = 0;
int runonce_outer = 0;
int runonce_middle = 0;
int runonce_inner = 0;
int runonce_trans = 0;
int notifyonce_outer = 1;
int notifyonce_trans = 1;

int calibration_process = 0;
int couple_process = 0;
int instrexchange_process = 0;
int homing_process = 0;
int block_calibration = 0;
uint32_t block_time;
uint32_t previous_block_time;
int block_interval = 10000;

bool outer_calibrated = false;
bool trans_calibrated = false;

bool firstval_outer = true;
bool firstval_middle = true;
bool firstval_inner = true;
bool firstval_trans = true;
bool firstval = true;

bool reverse_tip_stall = true;


////////////// VARIABLES FOR CALIBRATION CYCLE AND POSITION TRACKING ///////////////
uint32_t timecal_outer = 0;
uint32_t timecal_trans = 0;
uint32_t timepos = 0;
uint32_t previous_cal_time_trans = 0;
uint32_t previous_cal_time_outer = 0;


uint32_t timecouple;
uint32_t previous_couple_time = 0;
const int couple_interval = 175;

int k = 0;  //counter for backlash position correction after homing outer shaft
int k_current = 0;

uint32_t timehome;
uint32_t previous_hom_time = 0;
int hom_time_interval = 75;

bool home_outer_dir = true;
bool home_middle_dir = true;
uint32_t previous_pos_time = 0;
const int calInterval = 20;
const int posInterval = 20;
int endpoint_outer = 0;
int endpoint_trans = 0;
int home_point_outer = 0;
int home_point_middle = 0;
int cal_counter_outer = 0;
int cal_counter_trans = 0;
int trans_started = 0;

int current_outer = 0;           //current position of outer shaft motor
int current_trans = 0;           //current position of trans shaft motor
int previous_current_outer = 0;  //previous position of outer shaft motor
int previous_current_trans = 0;  //previous position of transporter shaft motor
bool mid_pos_trans = false;
bool mid_pos_outer = false;

bool coupled = false;

bool outer_home_return = false;
bool middle_home_return = false;

// Serial communication
#define SERIAL_PORT_1 Serial1

#define R_SENSE 0.11f  // R_SENSE for current calc.

#define driver_ADDRESS 0b00    // TMC2209 driver_outer address according to MS1 (LOW) and MS2 (LOW)
#define driver_ADDRESS_1 0b10  // TMC2209 driver_middle address according to MS1 (HIGH) and MS2 (LOW)
#define driver_ADDRESS_2 0b01  // TMC2209 driver_inner address according to MS1 (LOW) and MS2 (HIGH)
#define driver_ADDRESS_3 0b11  // TMC2209 driver_trans address according to MS1 (HIGH) and MS2 (HIGH)


TMC2209Stepper driver_outer(&SERIAL_PORT_1, R_SENSE, driver_ADDRESS);
TMC2209Stepper driver_middle(&SERIAL_PORT_1, R_SENSE, driver_ADDRESS_1);
TMC2209Stepper driver_inner(&SERIAL_PORT_1, R_SENSE, driver_ADDRESS_2);
TMC2209Stepper driver_trans(&SERIAL_PORT_1, R_SENSE, driver_ADDRESS_3);

/////////////// SET UP ///////////////
void setup() {
  Serial.begin(115200);
  SERIAL_PORT_1.begin(115200);

  pinMode(X_ENABLE_PIN, OUTPUT);  //X pins are connected to outer shaft motor
  digitalWrite(X_ENABLE_PIN, LOW);

  pinMode(Y_ENABLE_PIN, OUTPUT);  //Y pins are connected to transporter motor
  digitalWrite(Y_ENABLE_PIN, LOW);

  pinMode(Z_ENABLE_PIN, OUTPUT);  //Z pins are connected to inner rod motor
  digitalWrite(Z_ENABLE_PIN, LOW);

  pinMode(E0_ENABLE_PIN, OUTPUT);  //E0 pins are connected to middle shaft motor
  digitalWrite(E0_ENABLE_PIN, LOW);

  setdrivers();

  radio.begin();
  radio.openReadingPipe(0, address);
  radio.setPALevel(RF24_PA_MIN);
  radio.startListening();
}

/////////////// LOOP START ///////////////
void loop() {

  /////////////// RADIO IMPORT ///////////////
  if (radio.available()) {
    radio_import();
  }
  radio_check();

  /////////////// CALIBRATION NOTIFICATION ///////////////
  if ((outer_calibrated == false || trans_calibrated == false) && (startBend_for == 1 || startBend_back == 1 || startTip_close == 1 || startTip_open == 1) && runonce == 0) {
    Serial.println("Calibration is required first");
    runonce = 1;
  }

  /////////////// MOTOR SEQUENCE OPTIONS (HOMING, CALIBRATION, INSTRUMENT EXCHANGE & COUPLING) ///////////////

  //HOMING CYCLE
  if (initializeHoming == 1 && (outer_homed == false || middle_homed == false || trans_homed == false)) {
    startHoming = 1;
    homing_process = 1;
    while (startHoming == 1) {
      Start_Homing();
    }
  }


  //CALIBRATION CYCLE
  if (initializeCalibration == 1 && calibration_process < 10) {
    startCalibration = 1;
    calibration_process = 1;
    while (startCalibration == 1) {
      Start_Calibration();
    }
  }

  //BLOCK CALIBRATION CYLE FIRST 10 SECONDS AFTER COMPLETION
  if (calibration_process == 10 && block_calibration == 1) {
    block_time = millis();
    if ((block_time - previous_block_time) > block_interval) {
      previous_block_time = block_time;
      calibration_process = 0;
      block_calibration = 0;
      Serial.println("Calibration unblocked");
    }
  }

  //INSTRUMENT EXCHANGE CYCLE
  if (initializeInstExchange == 1 && (outer_homed == false || middle_homed == false)) {
    startInstExchange = 1;
    instrexchange_process = 1;
    while (startInstExchange == 1) {
      Start_Instrument_Exchange();
    }
  }

  //COUPLING CYCLE
  if (initializeCouple == 1 && coupled == false) {
    startCouple = 1;
    couple_process = 1;
    while (startCouple == 1) {
      Start_Couple();
    }
  }

  /////////////// MOTOR CONTROLLING (BENDING FORWARDS & BACKWARDS, ROTATE INSTRUMENT CW & CCW, ROTATE TIP CW & CCW, OPEN & CLOSE TIP) ///////////////
  // BEND INSTRUMENT FORWARD
  if (startBend_for == 1 && outer_calibrated == true && (current_outer < (cal_counter_outer - 150))) {
    if (runonce == 0) {
      Serial.println("Bend forward");
      driver_outer.shaft(false);
      driver_outer.VACTUAL(6000);

      reset_home();
      previous_pos_time = millis();
      runonce = 1;
      timepos = 0;
    }

    timepos = millis();

    if ((timepos - previous_pos_time) > posInterval) {
      // Run position counter every 0.05s
      previous_pos_time = timepos;
      current_outer = current_outer + 6;
      Serial.println(current_outer);
    }
  }

  // BEND INSTRUMENT BACKWARD
  if (startBend_back == 1 && outer_calibrated == true && current_outer > 0) {
    if (runonce == 0) {
      Serial.println("Bend backward");
      driver_outer.shaft(true);
      driver_outer.VACTUAL(6000);

      reset_home();
      previous_pos_time = millis();
      runonce = 1;
      timepos = 0;
    }

    timepos = millis();

    if ((timepos - previous_pos_time) > posInterval) {
      // Run position counter every 0.05s
      previous_pos_time = timepos;
      current_outer = current_outer - 6;
      Serial.println(current_outer);
    }
  }

  // CLOSE TIP
  if (startTip_close == 1 && trans_calibrated == true && current_trans > 50) {
    if (reverse_tip_stall == true && runonce == 0) {
      Serial.println("REVERSE TIP DIRECTION");
      runonce = 1;
    } else if (reverse_tip_stall == false) {
      if (runonce == 0) {
        Serial.println("Close tip");
        driver_trans.shaft(true);
        driver_trans.VACTUAL(8000);
        previous_pos_time = millis();
        previous_load_measuring_time = millis();
        ms = millis();
        runonce = 1;
        timepos = 0;
        accumulated_load_trans = 0;
        firstval_trans = true;
        load_measures_counter = 0;
      }

      timepos = millis();

      if ((timepos - previous_pos_time) > posInterval) {
        // Run position counter every 0.05s
        previous_pos_time = timepos;
        current_trans = current_trans - 8;
        Serial.println(current_trans);
      }

      stall_trans_safety_detection();
    }
  }

  // OPEN TIP
  if (startTip_open == 1 && trans_calibrated == true && current_trans < (cal_counter_trans - 200)) {
    if (runonce == 0) {
      Serial.println("Open tip");
      driver_trans.shaft(false);
      driver_trans.VACTUAL(8000);
      previous_pos_time = millis();
      runonce = 1;
      timepos = 0;
      reverse_tip_stall = false;
    }

    timepos = millis();

    if ((timepos - previous_pos_time) > posInterval) {
      // Run position counter every 0.05s
      previous_pos_time = timepos;
      current_trans = current_trans + 8;
      Serial.println(current_trans);
    }
  }

  // ROTATE END EFFECTOR CW
  if (startRotEnd_CW == 1 && runonce == 0) {
    Serial.println("Rotate end effector CW");
    driver_inner.shaft(false);
    driver_trans.shaft(false);

    driver_inner.VACTUAL(6000);
    driver_trans.VACTUAL(6000);
    runonce = 1;
  }

  // ROTATE END EFFECTOR CCW
  if (startRotEnd_CCW == 1 && runonce == 0) {
    Serial.println("Rotate end effector CCW");
    driver_inner.shaft(true);
    driver_trans.shaft(true);

    driver_inner.VACTUAL(6000);
    driver_trans.VACTUAL(6000);
    runonce = 1;
  }

  // ROTATE INSTRUMENT CW
  if (startRotate_CW == 1 && runonce == 0) {
    Serial.println("Rotate instrument CW");
    driver_inner.shaft(false);
    driver_trans.shaft(false);
    driver_outer.shaft(false);
    driver_middle.shaft(false);

    driver_inner.VACTUAL(6000);
    driver_trans.VACTUAL(6000);
    driver_outer.VACTUAL(6000);
    driver_middle.VACTUAL(6000);
    reset_home();
    runonce = 1;
  }

  // ROTATE INSTRUMENT CCW
  if (startRotate_CCW == 1 && runonce == 0) {
    Serial.println("Rotate instrument CCW");
    driver_inner.shaft(true);
    driver_trans.shaft(true);
    driver_outer.shaft(true);
    driver_middle.shaft(true);

    driver_inner.VACTUAL(6000);
    driver_trans.VACTUAL(6000);
    driver_outer.VACTUAL(6000);
    driver_middle.VACTUAL(6000);

    reset_home();
    runonce = 1;
  }

  if (current_trans != previous_current_trans) {
    previous_current_trans = current_trans;
    notifyonce_trans = 0;
  }

  if (current_outer != previous_current_outer) {
    previous_current_outer = current_outer;
    notifyonce_outer = 0;
  }

  // NOTIFY WHEN ENDPOINT REACHED WHEN BENDING
  if ((current_outer >= (cal_counter_outer - 150) || current_outer <= 0) && outer_calibrated == true && notifyonce_outer == 0) {
    driver_outer.VACTUAL(0);
    Serial.println("endpoint reached, reverse bending direction");
    notifyonce_outer = 1;
  }

  // NOTIFY WHEN ENDPOINT REACHED WHEN OPENING/CLOSING TIP
  if ((current_trans >= (cal_counter_trans - 200) || current_trans <= 50) && trans_calibrated == true && notifyonce_trans == 0) {
    driver_trans.VACTUAL(0);
    Serial.println("endpoint reached, reverse direction for tip");
    notifyonce_trans = 1;
  }

  // STOP ACTUATION WHEN OPERATOR RELEASES BUTTON
  if (startRotate_CW == 0 && startRotate_CCW == 0 && startBend_for == 0 && startBend_back == 0 && startTip_close == 0 && startTip_open == 0 && startRotEnd_CCW == 0 && startRotEnd_CW == 0) {
    driver_outer.VACTUAL(0);
    driver_middle.VACTUAL(0);
    driver_inner.VACTUAL(0);
    driver_trans.VACTUAL(0);
  }
}

////////////// READ AND RECEIVE RADIO INFORMATION FROM ARDUINO CONTROLLER ///////////////
void radio_import() {
  radio.read(&data, sizeof(data));
  handshake_check = data.Handshake;
  // Process the received data

  if (data.E_stop_button == 1) {
    //stop all motors
    driver_outer.VACTUAL(0);
    driver_middle.VACTUAL(0);
    driver_inner.VACTUAL(0);
    driver_trans.VACTUAL(0);
    Serial.println("Emergency stop");
    //reset all parameters
    reset_parameters();

    //wait for 3 seconds
    Serial.println("wait 3 seconds");
    delay(3000);
  }

  if (data.rotate_CW != startRotate_CW || data.rotate_CCW != startRotate_CCW || data.bend_for != startBend_for || data.bend_back != startBend_back || data.Up_button != startTip_close || data.Down_button != startTip_open || data.Left_button != startRotEnd_CCW || data.Right_button != startRotEnd_CW || data.E_button != initializeCalibration || data.F_button != initializeHoming || data.Joy_button != initializeInstExchange || data.Extra_button != initializeCouple) {
    startRotate_CW = data.rotate_CW;
    startRotate_CCW = data.rotate_CCW;
    startBend_for = data.bend_for;
    startBend_back = data.bend_back;
    startTip_close = data.Up_button;
    startTip_open = data.Down_button;
    startRotEnd_CCW = data.Left_button;
    startRotEnd_CW = data.Right_button;

    initializeCalibration = data.E_button;
    initializeHoming = data.F_button;
    initializeInstExchange = data.Joy_button;
    initializeCouple = data.Extra_button;

    val = 0;
    load_measures_counter = 0;
    accumulated_load = 0;
    runonce = 0;
    coupled = false;

    if (initializeInstExchange == 0) {
      outer_homed = false;
      middle_homed = false;
    }

    Serial.print(data.Up_button);
    Serial.print(data.Right_button);
    Serial.print(data.Down_button);
    Serial.print(data.Left_button);
    Serial.print(data.E_button);
    Serial.print(data.F_button);
    Serial.print(data.Joy_button);
    Serial.print(data.Extra_button);
    Serial.print(data.rotate_CW);
    Serial.print(data.rotate_CCW);
    Serial.print(data.bend_for);
    Serial.print(data.bend_back);
    Serial.println("\t received");
  }
}

////////////// COMMUNICATION WITH ARDUINO CONTROLLER LOST!! ///////////////
void radio_check() {
  currentTime = millis();

  if (handshake_check != prev_handshake_check_state) {
    prev_handshake_check_state = handshake_check;
    lastHandshakeTime = currentTime;  // Update lastHandshakeTime
    if (flag == 1) {
      Serial.println("COMMUNICATION WITH CONTROLLER RESTORED!");
      flag = 0;
    }
  }

  // Communication FAILED
  if (currentTime - lastHandshakeTime >= handshakeInterval && flag == 0) {
    driver_outer.VACTUAL(0);
    driver_middle.VACTUAL(0);
    driver_inner.VACTUAL(0);
    driver_trans.VACTUAL(0);

    Serial.println("NO COMMUNICATION WITH CONTROLLER! MOTORS STAND STILL!");

    reset_parameters();
  }
}

////////////// HOMING CYCLE WITH HALL SENSOR /////////////// (homing and instrument exchange combined)
void Start_Homing() {

  /////////////// RADIO IMPORT ///////////////
  if (radio.available()) {
    radio_import();
  }
  radio_check();


  // activate motors
  if (homing_process == 1) {
    Serial.println("homing started");

    driver_outer.shaft(true);
    driver_middle.shaft(true);

    driver_outer.VACTUAL(6000);
    driver_middle.VACTUAL(6000);

    driver_trans.shaft(true);
    driver_trans.VACTUAL(8000);


    outer_homed = false;
    middle_homed = false;
    trans_homed = false;
    hall_threshold_outer = false;
    hall_threshold_middle = false;
    home_outer_dir = true;
    home_middle_dir = true;
    previous_load_measuring_time = millis();
    homing_process = 2;
  }

  // check hall sensor
  if (hall_threshold_outer == false) {
    hall_outer = analogRead(HALL_OUTER_PIN);  // Read the analog value from the Hall sensor
    if (hall_outer > 681) {
      hall_threshold_outer = true;
      driver_outer.VACTUAL(0);
      driver_inner.VACTUAL(0);
    }
  }

  if (hall_threshold_middle == false) {
    hall_middle = analogRead(HALL_MIDDLE_PIN);  // Read the analog value from the Hall sensor
    if (hall_middle > 1004) {
      hall_threshold_middle = true;
      driver_middle.VACTUAL(0);
    }
  }

  // Hall sensors outer and middle reached threshold - start home correction
  if (hall_threshold_outer == true && hall_threshold_middle == true && (outer_homed == false || middle_homed == false) && trans_homed == true) {
    start_home_correction();
  }

  //Stallguard readings trans (to detect end-stop) and outer (to detect end stop if middle is already homed) and middle (to detect end stop if outer is already homed)
  // MEASURING STALLGUARD VALUES
  ms = millis();

  if ((ms - previous_load_measuring_time) > LoadMeasureInterval) {
    // Run load measuring every 0.05s
    previous_load_measuring_time = ms;
    load_measures_counter++;

    // ACTIVATE STALLGUARD ONLY when one of the shafts is homed already

    if (trans_homed == false) {
      accumulated_load_trans += driver_trans.SG_RESULT();  // Actual load measuring
    }
    if (hall_threshold_middle == true && hall_threshold_outer == false) {
      accumulated_load_outer += driver_outer.SG_RESULT();  // Actual load measuring
    }
    if (hall_threshold_outer == true && hall_threshold_middle == false) {
      accumulated_load_middle += driver_middle.SG_RESULT();  // Actual load measuring
    }
  }


  if (load_measures_counter >= LoadMeasureCount) {
    // Calculate the average load value
    if (trans_homed == false && endpoint_trans == 0) {
      if (firstval_trans == false) {
        val_trans = accumulated_load_trans / load_measures_counter;
        accumulated_load_trans = 0;
        Serial.print("SG Value Transporter \t");
        Serial.println(val_trans);
      } else if (firstval_trans == true) {
        firstval_trans = false;
        accumulated_load_trans = 0;
        val_trans = 0;
      }
    }

    if (hall_threshold_middle == true && hall_threshold_outer == false) {
      if (firstval_outer == false) {
        val_outer = accumulated_load_outer / load_measures_counter;
        accumulated_load_outer = 0;
        Serial.print("\t SG Value Outer \t");
        Serial.println(val_outer);
      } else if (firstval_outer == true) {
        firstval_outer = false;
        accumulated_load_outer = 0;
        val_outer = 0;
      }
    }

    if (hall_threshold_outer == true && hall_threshold_middle == false) {
      if (firstval_middle) {
        firstval_middle = false;
        accumulated_load_middle = 0;
        val_middle = 0;
      } else {
        val_middle = accumulated_load_middle / load_measures_counter;
        accumulated_load_middle = 0;
        Serial.print("\t SG Value Middle \t");
        Serial.println(val_middle);
      }
    }
    load_measures_counter = 0;
  }

  // detect 1 endstop based on calculated stallguard value OUTER SHAFT
  if (val_outer <= 160 && val_outer > 0 && home_point_outer == 0) {
    driver_outer.VACTUAL(0);
    driver_middle.VACTUAL(0);
    driver_trans.VACTUAL(0);
    driver_inner.VACTUAL(0);
    Serial.println("Stop, instrument still connected. Homing failed.");
    delay(100);
    val_outer = 0;
    home_point_outer = 1;
    runonce_outer = 0;
    firstval_outer = true;
  }

  // detect 1 endstop based on calculated stallguard value MIDDLE SHAFT
  if (val_middle <= 95 && val_middle > 0 && home_point_middle == 0) {
    driver_outer.VACTUAL(0);
    driver_middle.VACTUAL(0);
    driver_trans.VACTUAL(0);
    driver_inner.VACTUAL(0);
    Serial.println("Stop, instrument still connected. Homing failed.");
    delay(100);
    val_middle = 0;
    home_point_middle = 1;
    runonce_middle = 0;
    firstval_middle = true;
  }


  // stallguard is triggered twice and home position is not reached - stop homing
  if (home_point_outer == 1 || home_point_middle == 1) {  //reverse homing direction
    Serial.println("Homing failed.");
    driver_outer.VACTUAL(0);
    driver_middle.VACTUAL(0);
    driver_trans.VACTUAL(0);
    driver_inner.VACTUAL(0);
    startHoming = 0;
    startInstExchange = 0;

    runonce_outer = 0;
    runonce_middle = 0;
    runonce = 0;

    home_point_outer = 0;
    home_point_middle = 0;

    val_outer = 0;
    val_middle = 0;

    hall_threshold_outer = false;
    hall_threshold_middle = false;

    home_outer_dir = true;
    home_middle_dir = true;

    firstval_outer = true;
    firstval_middle = true;
    firstval_trans = true;
    homing_process = 0;
  }

  // DETECT ENDPOINT OF TRANSPORTER
  if (val_trans <= 200 && val_trans > 0 && endpoint_trans == 0) {
    driver_trans.VACTUAL(0);
    Serial.println("Endpoint transporter reached");
    delay(100);
    val_trans = 0;
    endpoint_trans = 1;
    cal_counter_trans = 0;
    timecal_trans = 0;
    runonce_trans = 0;
  }

  //CHANGE DIRECTION AND BRING TRANSPORTER TO MID/HOME POSITION
  if (endpoint_trans == 1) {
    timecal_trans = millis();

    if (runonce_trans == 0) {
      driver_trans.shaft(false);
      driver_trans.VACTUAL(8000);
      runonce_trans = 1;
    }

    if ((timecal_trans - previous_cal_time_trans) > calInterval) {
      // Run position counter every 0.05s
      previous_cal_time_trans = timecal_trans;
      cal_counter_trans = cal_counter_trans + 8;
    }

    if (cal_counter_trans >= 1900) {
      driver_trans.VACTUAL(0);
      Serial.println("Home position transporter reached");
      val_trans = 0;
      endpoint_trans = 0;
      timecal_trans = 0;
      runonce_trans = 0;
      trans_homed = true;
    }
  }

  // homing is finished, reset all flags
  if (outer_homed == true && middle_homed == true && trans_homed == true) {

    runonce_outer = 1;
    runonce_middle = 1;
    runonce = 0;

    home_point_outer = 0;
    home_point_middle = 0;

    val_outer = 0;
    val_middle = 0;

    hall_threshold_outer = false;
    hall_threshold_middle = false;

    outer_home_return = false;
    middle_home_return = false;

    firstval_outer = true;
    firstval_middle = true;
    firstval_trans = true;

    outer_calibrated = false;
    trans_calibrated = false;

    k = 0;
    homing_process = 0;
  }
}

////////////// HOMING CYCLE CORRECTION FOR PERFECT INSTRUMENT DISASSEMBLY POSITION ///////////////
void start_home_correction() {
  start_inst_exchange_correction();
}

////////////// CALIBRATION CYCLE STALL GUARD FOR TRANSPORTER SHAFT MEASUREMENT ///////////////
void Start_Calibration() {

  /////////////// RADIO IMPORT ///////////////
  if (radio.available()) {
    radio_import();
  }
  radio_check();

  //Start running OUTER SHAFT first
  if (calibration_process == 1) {
    Serial.println("calibration outer started");
    driver_outer.shaft(true);
    driver_outer.VACTUAL(6000);

    reset_home();
    previous_load_measuring_time = millis();

    calibration_process = 2;
  }

  // MEASURING STALLGUARD VALUES
  ms = millis();

  if ((ms - previous_load_measuring_time) > LoadMeasureInterval) {
    // Run load measuring every 0.05s
    previous_load_measuring_time = ms;
    load_measures_counter++;
    if (calibration_process > 5) {
      accumulated_load_trans += driver_trans.SG_RESULT();  // Actual load measuring
    }
    if (calibration_process < 6) {
      accumulated_load_outer += driver_outer.SG_RESULT();  // Actual load measuring
    }
  }

  if (load_measures_counter >= LoadMeasureCount) {
    // Calculate the average load value

    if (calibration_process == 2 || calibration_process == 3) {
      if (firstval_outer == false) {
        val_outer = accumulated_load_outer / load_measures_counter;
        accumulated_load_outer = 0;
        Serial.print("\t SG Value Outer \t");
        Serial.println(val_outer);
      } else if (firstval_outer == true) {
        firstval_outer = false;
        accumulated_load_outer = 0;
        val_outer = 0;
        Serial.println("first val outer deleted");
      }
    }

    if (calibration_process == 6 || calibration_process == 7) {
      if (firstval_trans == false) {
        val_trans = accumulated_load_trans / load_measures_counter;
        accumulated_load_trans = 0;
        Serial.print("SG Value Transporter \t");
        Serial.println(val_trans);
      } else if (firstval_trans == true) {
        firstval_trans = false;
        accumulated_load_trans = 0;
        val_trans = 0;
        Serial.println("first val trans deleted");
      }
    }

    load_measures_counter = 0;
  }

  // DETECT FIRST ENDPOINT OF OUTER SHAFT
  if (val_outer <= 160 && val_outer > 0 && calibration_process == 2) {
    driver_outer.VACTUAL(0);
    Serial.println("Stop, first endpoint outer reached");
    delay(100);
    val_outer = 0;
    cal_counter_outer = 0;
    timecal_outer = 0;
    firstval_outer = true;

    calibration_process = 3;
  }

  // CHANGE DIRECTION AND SEARCH 2nd ENDPOINT (and distance)
  if (calibration_process == 3) {
    timecal_outer = millis();

    if (runonce == 0) {
      driver_outer.shaft(false);
      driver_outer.VACTUAL(6000);

      previous_load_measuring_time = millis();
      runonce = 1;
    }

    if ((timecal_outer - previous_cal_time_outer) > calInterval) {
      // Run position counter every 0.05s
      previous_cal_time_outer = timecal_outer;
      cal_counter_outer = cal_counter_outer + 6;
      Serial.println(cal_counter_outer);
    }

    if (val_outer <= 160 && val_outer > 0) {
      driver_outer.VACTUAL(0);
      Serial.println("Stop, second endpoint outer detected");
      Serial.println("position range is:");
      Serial.println(cal_counter_outer);
      val_outer = 0;
      timecal_outer = 0;
      current_outer = cal_counter_outer - 20;
      previous_current_outer = current_outer;
      firstval_outer = true;

      runonce = 0;
      calibration_process = 4;
    }
  }

  // AFTER OUTER IS CALIBRATED, FIRST CENTER POSITION OUTER
  if (calibration_process == 4) {
    Serial.println("centering outer actuator");
    delay(100);
    mid_position_outer();
  }

  // AFTER OUT IS CENTERED, START CALIBRATION OF TRANS
  if (calibration_process == 5) {
    Serial.println("calibration trans started");
    driver_trans.shaft(true);
    driver_trans.VACTUAL(8000);
    previous_load_measuring_time = millis();

    calibration_process = 6;
  }

  // DETECT FIRST ENDPOINT OF TRANSPORTER
  if (val_trans <= 190 && val_trans > 0 && calibration_process == 6) {
    driver_trans.VACTUAL(0);
    Serial.println("Stop, first endpoint transporter reached");
    delay(100);
    val_trans = 0;
    cal_counter_trans = 0;
    timecal_trans = 0;
    firstval_trans = true;

    calibration_process = 7;
  }

  // CHANGE DIRECTION AND SEARCH 2nd ENDPOINT (and distance)
  if (calibration_process == 7) {
    timecal_trans = millis();

    if (runonce == 0) {
      driver_trans.shaft(false);
      driver_trans.VACTUAL(8000);
      previous_load_measuring_time = millis();
      runonce = 1;
    }

    if ((timecal_trans - previous_cal_time_trans) > calInterval) {
      previous_cal_time_trans = timecal_trans;
      cal_counter_trans = cal_counter_trans + 8;
    }

    if (val_trans <= 190 && val_trans > 0) {
      driver_trans.VACTUAL(0);
      Serial.println("Stop, second endpoint transporter detected");
      Serial.println("position range is:");
      Serial.println(cal_counter_trans);
      val_trans = 0;
      timecal_trans = 0;
      current_trans = cal_counter_trans - 20;
      previous_current_trans = current_trans;
      firstval_trans = true;

      runonce = 0;
      calibration_process = 8;
    }
  }

  if (calibration_process == 8) {
    Serial.println("Centering transporter");
    delay(500);
    mid_position_trans();
  }

  if (calibration_process == 9) {
    Serial.println("Calibration complete");
    outer_calibrated = true;
    trans_calibrated = true;
    calibration_process = 10;
    block_calibration = 1;
    startCalibration = 0;
    previous_block_time = millis();
  }
}

////////////// RETURNING TO CENTER POSITION AFTER CALIBRATION CYCLE OUTER ///////////////
void mid_position_outer() {

  while (calibration_process == 4) {

    /////////////// RADIO IMPORT ///////////////
    if (radio.available()) {
      radio_import();
    }
    radio_check();

    if (runonce == 0) {
      driver_outer.shaft(true);
      driver_outer.VACTUAL(6000);
      runonce = 1;
    }

    timepos = millis();

    if ((timepos - previous_pos_time) > posInterval) {
      previous_pos_time = timepos;
      current_outer = current_outer - 6;
      Serial.println(current_outer);
    }

    if (current_outer <= ((cal_counter_outer / 2) - 100)) {
      driver_outer.VACTUAL(0);

      runonce = 0;
      calibration_process = 5;
    }
  }
}

////////////// RETURNING TO CENTER POSITION AFTER CALIBRATION CYCLE TRANSPORTER ///////////////
void mid_position_trans() {

  while (calibration_process == 8) {

    /////////////// RADIO IMPORT ///////////////
    if (radio.available()) {
      radio_import();
    }
    radio_check();

    if (runonce == 0) {
      driver_trans.shaft(true);
      driver_trans.VACTUAL(8000);
      runonce = 1;
    }

    timepos = millis();

    if ((timepos - previous_pos_time) > posInterval) {
      previous_pos_time = timepos;

      if (mid_pos_trans == false) {
        current_trans = current_trans - 8;
        Serial.println(current_trans);
      }
    }

    if (current_trans <= 200) {
      driver_trans.VACTUAL(0);

      runonce = 0;
      calibration_process = 9;
    }
  }
}


////////////// SEQUENCE TO PREPARE FOR INSTRUMENT EXCHANGE ///////////////
void Start_Instrument_Exchange() {

  /////////////// RADIO IMPORT ///////////////
  if (radio.available()) {
    radio_import();
  }
  radio_check();


  // activate motors
  if (instrexchange_process == 1) {
    Serial.println("Instrument exchange procedure started");
    driver_outer.shaft(true);
    driver_middle.shaft(true);
    driver_trans.shaft(true);
    driver_inner.shaft(true);

    driver_outer.VACTUAL(6000);
    driver_middle.VACTUAL(6000);
    driver_trans.VACTUAL(6000);
    driver_inner.VACTUAL(6000);

    outer_homed = false;
    middle_homed = false;
    hall_threshold_outer = false;
    hall_threshold_middle = false;
    instrexchange_process = 2;
    home_outer_dir = true;
    home_middle_dir = true;
    previous_load_measuring_time = millis();
  }

  // check hall sensor
  if (hall_threshold_outer == false) {
    hall_outer = analogRead(HALL_OUTER_PIN);  // Read the analog value from the Hall sensor
    if (hall_outer > 681) {
      hall_threshold_outer = true;
      driver_outer.VACTUAL(0);
      driver_trans.VACTUAL(0);
      driver_inner.VACTUAL(0);
    }
  }

  if (hall_threshold_middle == false) {
    hall_middle = analogRead(HALL_MIDDLE_PIN);  // Read the analog value from the Hall sensor
    if (hall_middle > 1004) {
      hall_threshold_middle = true;
      driver_middle.VACTUAL(0);
      driver_trans.VACTUAL(0);
      driver_inner.VACTUAL(0);
    }
  }

  // Hall sensors outer and middle reached threshold - start home correction
  if (hall_threshold_outer == true && hall_threshold_middle == true && (outer_homed == false || middle_homed == false)) {
    start_inst_exchange_correction();
  }


  //Stallguard readings outer (to detect end stop if middle is already homed) and middle (to detect end stop if outer is already homed)
  // MEASURING STALLGUARD VALUES
  ms = millis();

  if ((ms - previous_load_measuring_time) > LoadMeasureInterval) {
    // Run load measuring every 0.05s
    previous_load_measuring_time = ms;
    load_measures_counter++;

    // ACTIVATE STALLGUARD ONLY when one of the shafts is homed already
    if (hall_threshold_middle == true && hall_threshold_outer == false) {
      accumulated_load_outer += driver_outer.SG_RESULT();  // Actual load measuring
    }
    if (hall_threshold_outer == true && hall_threshold_middle == false) {
      accumulated_load_middle += driver_middle.SG_RESULT();  // Actual load measuring
    }
  }


  if (load_measures_counter >= LoadMeasureCount) {
    // Calculate the average load value

    if (hall_threshold_middle == true && hall_threshold_outer == false) {
      if (firstval_outer == false) {
        val_outer = accumulated_load_outer / load_measures_counter;
        accumulated_load_outer = 0;
        Serial.print("\t SG Value Outer \t");
        Serial.println(val_outer);
      } else if (firstval_outer == true) {
        firstval_outer = false;
        accumulated_load_outer = 0;
        val_outer = 0;
      }
    }

    if (hall_threshold_outer == true && hall_threshold_middle == false) {
      if (firstval_middle) {
        firstval_middle = false;
        accumulated_load_middle = 0;
        val_middle = 0;
      } else {
        val_middle = accumulated_load_middle / load_measures_counter;
        accumulated_load_middle = 0;
        Serial.print("\t SG Value Middle \t");
        Serial.println(val_middle);
      }
    }
    load_measures_counter = 0;
  }

  // detect 1 endstop based on calculated stallguard value OUTER SHAFT
  if (val_outer < 160 && val_outer > 0 && home_point_outer == 0) {
    driver_outer.VACTUAL(0);
    driver_trans.VACTUAL(0);
    driver_inner.VACTUAL(0);
    Serial.println("Stop, physical limit reached.");
    delay(100);
    val_outer = 0;
    home_point_outer = 1;
    runonce_outer = 0;
    firstval_outer = true;
  }

  // reverse direction of OUTER SHAFT after 1 endstop sensed by stallguard is triggered
  if (home_point_outer == 1 && hall_threshold_outer == false && runonce_outer == 0) {  //reverse homing direction
    Serial.println("outer homing direction reversed");
    driver_outer.shaft(false);
    driver_trans.shaft(false);
    driver_inner.shaft(false);

    driver_outer.VACTUAL(6000);
    driver_trans.VACTUAL(6000);
    driver_inner.VACTUAL(6000);

    runonce_outer = 1;
    home_outer_dir = false;
  }

  // detect 2 endstop based on calculated stallguard value OUTER SHAFT
  if (val_outer < 160 && val_outer > 0 && home_point_outer == 1) {
    driver_outer.VACTUAL(0);
    driver_trans.VACTUAL(0);
    driver_inner.VACTUAL(0);
    Serial.println("Stop, physical limit reached.");
    delay(100);
    val_outer = 0;
    home_point_outer = 2;
    runonce_outer = 0;
    firstval_outer = true;
  }

  // detect 1 endstop based on calculated stallguard value MIDDLE SHAFT
  if (val_middle <= 160 && val_middle > 0 && home_point_middle == 0) {
    driver_middle.VACTUAL(0);
    Serial.println("Stop, physical limit reached middle actuator.");
    delay(100);
    val_middle = 0;
    home_point_middle = 1;
    runonce_middle = 0;
    firstval_middle = true;
  }

  // reverse direction of MIDDLE SHAFT after 1 endstop sensed by stallguard is triggered
  if (home_point_middle == 1 && hall_threshold_middle == false && runonce_middle == 0) {  //reverse homing direction
    Serial.println("middle homing direction reversed");
    driver_middle.shaft(false);
    driver_middle.VACTUAL(6000);
    runonce_middle = 1;
    home_middle_dir = false;
  }

  // detect 2 endstop based on calculated stallguard value MIDDLE SHAFT
  if (val_middle <= 160 && val_middle > 0 && home_point_middle == 1) {
    driver_middle.VACTUAL(0);
    Serial.println("Stop, physical limit reached 2nd time middle actuator.");
    delay(100);
    val_middle = 0;
    home_point_middle = 2;
    runonce_middle = 0;
    firstval_middle = true;
  }

  // stallguard is triggered twice and home position is not reached - stop homing
  if ((home_point_outer == 2 || home_point_middle == 2) && (runonce_outer == 0 || runonce_middle == 0)) {  //reverse homing direction
    Serial.println("Instrument exchange failed.");
    driver_outer.VACTUAL(0);
    driver_middle.VACTUAL(0);
    driver_trans.VACTUAL(0);
    driver_inner.VACTUAL(0);
    startHoming = 0;
    startInstExchange = 0;

    runonce_outer = 0;
    runonce_middle = 0;
    runonce = 0;

    home_point_outer = 0;
    home_point_middle = 0;

    val_outer = 0;
    val_middle = 0;

    hall_threshold_outer = false;
    hall_threshold_middle = false;

    home_outer_dir = true;
    home_middle_dir = true;

    firstval_outer = true;
    firstval_middle = true;
    firstval_trans = true;
    instrexchange_process = 0;
  }
}

////////////// INSTRUMENT EXCHANGE CYCLE CORRECTION FOR PERFECT INSTRUMENT DISASSEMBLY POSITION ///////////////
void start_inst_exchange_correction() {

  // time delay to spin motor a little further for absolute correct position - outer should be corrected in true and false shaft direction (with different correction margins)
  timehome = millis();

  if ((timehome - previous_hom_time) > hom_time_interval) {
    previous_hom_time = timehome;

    if (k == 0) {  // rotate half revolution clockwise
      Serial.println("home correction started");
      delay(500);
      driver_outer.shaft(false);
      driver_middle.shaft(false);
      driver_trans.shaft(false);
      driver_inner.shaft(false);

      driver_outer.VACTUAL(6000);
      driver_middle.VACTUAL(6000);
      driver_trans.VACTUAL(6000);
      driver_inner.VACTUAL(6000);
    }

    k++;

    if (k == 30) {  // return to inst exchange / home position 2nd time
      Serial.println("stop");
      driver_outer.VACTUAL(0);
      driver_middle.VACTUAL(0);
      driver_trans.VACTUAL(0);
      driver_inner.VACTUAL(0);
    }

    if (k == 32) {  // return to inst exchange / home position 2nd time
      Serial.println("return to home");
      driver_outer.shaft(true);
      driver_middle.shaft(true);
      driver_trans.shaft(true);
      driver_inner.shaft(true);

      driver_outer.VACTUAL(6000);
      driver_middle.VACTUAL(6000);
      driver_trans.VACTUAL(6000);
      driver_inner.VACTUAL(6000);
    }

    // read when hall sensor threshold is reached
    if (k > 30 && (outer_home_return == false || middle_home_return == false)) {
      hall_outer = analogRead(HALL_OUTER_PIN);    // Read the analog value from the Hall sensor
      hall_middle = analogRead(HALL_MIDDLE_PIN);  // Read the analog value from the Hall sensor
      if (hall_outer > 681 && outer_home_return == false) {
        driver_outer.VACTUAL(0);
        driver_trans.VACTUAL(0);
        driver_inner.VACTUAL(0);
        outer_home_return = true;
        k_current = k;
      }

      if (hall_middle > 1004 && middle_home_return == false) {
        driver_middle.VACTUAL(0);
        driver_trans.VACTUAL(0);
        driver_inner.VACTUAL(0);
        middle_home_return = true;
        k_current = k;
      }
    }

    // set last correction cyclus for perfect positioning
    if (outer_home_return == true && middle_home_return == true) {
      if (k == k_current) {  //outer shaft
        driver_outer.shaft(true);
        driver_middle.shaft(true);

        driver_outer.VACTUAL(4000);
        driver_middle.VACTUAL(4000);
      }

      if (k == k_current + 3) {  //outer shaft
        driver_outer.VACTUAL(0);
        outer_homed = true;
      }

      if (k == k_current + 3) {  //middle shaft
        driver_middle.VACTUAL(0);
        middle_homed = true;
      }
    }


    // homing/instrument exchange is finished, reset all flags
    if (outer_homed == true && middle_homed == true) {
      Serial.println("Instrument exchange position complete");
      runonce_outer = 1;
      runonce_middle = 1;
      runonce = 0;

      home_point_outer = 0;
      home_point_middle = 0;

      val_outer = 0;
      val_middle = 0;

      hall_threshold_outer = false;
      hall_threshold_middle = false;

      outer_home_return = false;
      middle_home_return = false;

      firstval_outer = true;
      firstval_middle = true;
      firstval_trans = true;

      outer_calibrated = false;
      trans_calibrated = false;

      k = 0;
      instrexchange_process = 0;
      startInstExchange = 0;
    }
  }
}

////////////// SEQUENCE TO COUPLE MOTOR TO SBU AND GB ///////////////
void Start_Couple() {
  /////////////// RADIO IMPORT ///////////////
  if (radio.available()) {
    radio_import();
  }
  radio_check();

  // non-blocking delay till motor stops and reverses
  timecouple = millis();

  if (k == 0 && couple_process == 1) {
    Serial.println("Start coupling cyclus");
    driver_inner.shaft(true);
    driver_trans.shaft(true);
    driver_outer.shaft(true);
    driver_middle.shaft(true);

    driver_inner.VACTUAL(8000);
    driver_trans.VACTUAL(8000);
    driver_outer.VACTUAL(8000);
    driver_middle.VACTUAL(8000);

    reset_home();
    previous_couple_time = millis();
    previous_load_measuring_time = millis();
    ms = millis();
    accumulated_load_trans = 0;
    firstval_trans = true;
    firstval_outer = true;
    firstval_middle = true;
    firstval_inner = true;
    firstval = true;
    load_measures_counter = 0;
    couple_process = 2;
  }

  if ((timecouple - previous_couple_time) > couple_interval) {
    previous_couple_time = timecouple;
    k++;
  }

  if ((k > 0 && k < 20) || (k > 22 && k < 46)) {
    stall_all_motors_safety_detection();
  }

  if (k < 20 && reverse_tip_stall == false) {
    stall_trans_safety_detection();
  }

  // stop motors t = 1 seconds
  if (k == 20 && couple_process == 2) {
    Serial.println("Pause");
    driver_inner.VACTUAL(0);
    driver_trans.VACTUAL(0);
    driver_outer.VACTUAL(0);
    driver_middle.VACTUAL(0);

    couple_process = 3;

    accumulated_load_trans = 0;
    firstval_trans = true;
    firstval_outer = true;
    firstval_middle = true;
    firstval_inner = true;
    firstval = true;
    load_measures_counter = 0;
    val_outer = 0;
    val_trans = 0;
    val_middle = 0;
    val_inner = 0;
  }

  // reverse motors t = 2 seconds
  if (k == 22 && couple_process == 3) {
    Serial.println("Reverse coupling cyclus");
    driver_inner.shaft(false);
    driver_trans.shaft(false);
    driver_outer.shaft(false);
    driver_middle.shaft(false);

    driver_inner.VACTUAL(8000);
    driver_trans.VACTUAL(8000);
    driver_outer.VACTUAL(8000);
    driver_middle.VACTUAL(8000);

    couple_process = 4;
  }

  // end coupling cyclus t = 3 seconds
  if (k == 46 && couple_process == 4) {
    Serial.println("Coupling cyclus complete");
    driver_inner.VACTUAL(0);
    driver_trans.VACTUAL(0);
    driver_outer.VACTUAL(0);
    driver_middle.VACTUAL(0);

    couple_process = 0;
    startCouple = 0;
    coupled = true;
    k = 0;
  }
}

////////////// CONFIGURE DRIVER SETTINGS VIA UART ///////////////
void setdrivers() {
  // OUTER DRIVER SETTINGS
  driver_outer.begin();                       // Start all the UART communications functions behind the scenes
  driver_outer.toff(4);                       //For operation with StealthChop, this parameter is not used, but it is required to enable the motor. In case of operation with StealthChop only, any setting is OK
  driver_outer.blank_time(24);                //Recommended blank time select value
  driver_outer.I_scale_analog(true);          // Disbaled to use the extrenal current sense resistors
  driver_outer.internal_Rsense(false);        // Use the external Current Sense Resistors. Do not use the internal resistor as it can't handle high current.
  driver_outer.mstep_reg_select(true);        //Microstep resolution selected by MSTEP register and NOT from the legacy pins.
  driver_outer.microsteps(motor_microsteps);  //Set the number of microsteps. Due to the "MicroPlyer" feature, all steps get converterd to 256 microsteps automatically. However, setting a higher step count allows you to more accurately more the motor exactly where you want.
  driver_outer.TPWMTHRS(0);                   //DisableStealthChop PWM mode/ Page 25 of datasheet
  driver_outer.semin(0);                      // Turn off smart current control, known as CoolStep. It's a neat feature but is more complex and messes with StallGuard.
  driver_outer.shaft(true);                   // Set the shaft direction.
  driver_outer.en_spreadCycle(false);         // Disable SpreadCycle. We want StealthChop becuase it works with StallGuard.
  driver_outer.pdn_disable(true);             // Enable UART control
  driver_outer.index_step(true);              //Index pin output step pulses from internal pulse generator
  driver_outer.rms_current(set_current);
  driver_outer.SGTHRS(set_stall);
  driver_outer.TCOOLTHRS(set_tcools);


  // MIDDLE DRIVER SETTINGS
  driver_middle.begin();                       // Start all the UART communications functions behind the scenes
  driver_middle.toff(4);                       //For operation with StealthChop, this parameter is not used, but it is required to enable the motor. In case of operation with StealthChop only, any setting is OK
  driver_middle.blank_time(24);                //Recommended blank time select value
  driver_middle.I_scale_analog(true);          // Disbaled to use the extrenal current sense resistors
  driver_middle.internal_Rsense(false);        // Use the external Current Sense Resistors. Do not use the internal resistor as it can't handle high current.
  driver_middle.mstep_reg_select(true);        //Microstep resolution selected by MSTEP register and NOT from the legacy pins.
  driver_middle.microsteps(motor_microsteps);  //Set the number of microsteps. Due to the "MicroPlyer" feature, all steps get converterd to 256 microsteps automatically. However, setting a higher step count allows you to more accurately more the motor exactly where you want.
  driver_middle.TPWMTHRS(0);                   //DisableStealthChop PWM mode/ Page 25 of datasheet
  driver_middle.semin(0);                      // Turn off smart current control, known as CoolStep. It's a neat feature but is more complex and messes with StallGuard.
  driver_middle.shaft(true);                   // Set the shaft direction.
  driver_middle.en_spreadCycle(false);         // Disable SpreadCycle. We want StealthChop becuase it works with StallGuard.
  driver_middle.pdn_disable(true);             // Enable UART control
  driver_middle.index_step(true);              //Index pin output step pulses from internal pulse generator
  driver_middle.rms_current(set_current);
  driver_middle.SGTHRS(set_stall);
  driver_middle.TCOOLTHRS(set_tcools);


  // INNER DRIVER SETTINGS
  driver_inner.begin();                       // Start all the UART communications functions behind the scenes
  driver_inner.toff(4);                       //For operation with StealthChop, this parameter is not used, but it is required to enable the motor. In case of operation with StealthChop only, any setting is OK
  driver_inner.blank_time(24);                //Recommended blank time select value
  driver_inner.I_scale_analog(true);          // Disbaled to use the extrenal current sense resistors
  driver_inner.internal_Rsense(false);        // Use the external Current Sense Resistors. Do not use the internal resistor as it can't handle high current.
  driver_inner.mstep_reg_select(true);        //Microstep resolution selected by MSTEP register and NOT from the legacy pins.
  driver_inner.microsteps(motor_microsteps);  //Set the number of microsteps. Due to the "MicroPlyer" feature, all steps get converterd to 256 microsteps automatically. However, setting a higher step count allows you to more accurately more the motor exactly where you want.
  driver_inner.TPWMTHRS(0);                   //DisableStealthChop PWM mode/ Page 25 of datasheet
  driver_inner.semin(0);                      // Turn off smart current control, known as CoolStep. It's a neat feature but is more complex and messes with StallGuard.
  driver_inner.shaft(true);                   // Set the shaft direction.
  driver_inner.en_spreadCycle(false);         // Disable SpreadCycle. We want StealthChop becuase it works with StallGuard.
  driver_inner.pdn_disable(true);             // Enable UART control
  driver_inner.index_step(true);              //Index pin output step pulses from internal pulse generator
  driver_inner.rms_current(set_current);
  driver_inner.SGTHRS(set_stall);
  driver_inner.TCOOLTHRS(set_tcools);

  // INNER DRIVER SETTINGS
  driver_trans.begin();                       // Start all the UART communications functions behind the scenes
  driver_trans.toff(4);                       //For operation with StealthChop, this parameter is not used, but it is required to enable the motor. In case of operation with StealthChop only, any setting is OK
  driver_trans.blank_time(24);                //Recommended blank time select value
  driver_trans.I_scale_analog(true);          // Disbaled to use the extrenal current sense resistors
  driver_trans.internal_Rsense(false);        // Use the external Current Sense Resistors. Do not use the internal resistor as it can't handle high current.
  driver_trans.mstep_reg_select(true);        //Microstep resolution selected by MSTEP register and NOT from the legacy pins.
  driver_trans.microsteps(motor_microsteps);  //Set the number of microsteps. Due to the "MicroPlyer" feature, all steps get converterd to 256 microsteps automatically. However, setting a higher step count allows you to more accurately more the motor exactly where you want.
  driver_trans.TPWMTHRS(0);                   //DisableStealthChop PWM mode/ Page 25 of datasheet
  driver_trans.semin(0);                      // Turn off smart current control, known as CoolStep. It's a neat feature but is more complex and messes with StallGuard.
  driver_trans.shaft(true);                   // Set the shaft direction.
  driver_trans.en_spreadCycle(false);         // Disable SpreadCycle. We want StealthChop becuase it works with StallGuard.
  driver_trans.pdn_disable(true);             // Enable UART control
  driver_trans.index_step(true);              //Index pin output step pulses from internal pulse generator
  driver_trans.rms_current(set_current);
  driver_trans.SGTHRS(set_stall);
  driver_trans.TCOOLTHRS(set_tcools);


  driver_outer.VACTUAL(0);
  driver_middle.VACTUAL(0);
  driver_inner.VACTUAL(0);
  driver_trans.VACTUAL(0);
}

////////////// SETS HOMING TO FALSE ///////////////
void reset_home() {
  outer_homed = false;
  middle_homed = false;
}

/////////////// DETECT STALL TRANSPORTER - SAFETY FEATURE - ONLY ACTIVE WHEN TIP CLOSES ///////////////
void stall_trans_safety_detection() {
  ms = millis();

  if ((ms - previous_load_measuring_time) > LoadMeasureInterval) {
    // Run load measuring every 0.05s
    previous_load_measuring_time = ms;
    load_measures_counter++;
    accumulated_load_trans += driver_trans.SG_RESULT();  // Actual load measuring
  }

  if (load_measures_counter >= LoadMeasureCount) {
    // Calculate the average load value

    if (firstval_trans == false) {
      val_trans = accumulated_load_trans / load_measures_counter;
      accumulated_load_trans = 0;
      Serial.print("SG Value Transporter \t");
      Serial.println(val_trans);
    } else if (firstval_trans == true) {
      firstval_trans = false;
      accumulated_load_trans = 0;
      val_trans = 0;
    }

    load_measures_counter = 0;
  }
  if (val_trans <= 170 && val_trans > 0) {
    driver_trans.VACTUAL(0);
    driver_inner.VACTUAL(0);
    Serial.println("Stop, stall detected");
    delay(100);
    val_trans = 0;
    runonce_trans = 0;
    firstval_trans = true;
    reverse_tip_stall = true;
  }
}

/////////////// DETECT STALL ALL MOTORS DURING COUPLING SEQUENCE - SAFETY FEATURE  ///////////////
void stall_all_motors_safety_detection() {
  ms = millis();

  if ((ms - previous_load_measuring_time) > LoadMeasureInterval) {
    // Run load measuring every 0.05s
    previous_load_measuring_time = ms;
    load_measures_counter++;
    accumulated_load_trans += driver_trans.SG_RESULT();    // Actual load measuring
    accumulated_load_inner += driver_inner.SG_RESULT();    // Actual load measuring
    accumulated_load_middle += driver_middle.SG_RESULT();  // Actual load measuring
    accumulated_load_outer += driver_outer.SG_RESULT();    // Actual load measuring
  }

  if (load_measures_counter >= LoadMeasureCount) {
    // Calculate the average load value
    if (firstval == false) {
      val_trans = accumulated_load_trans / load_measures_counter;
      val_outer = accumulated_load_trans / load_measures_counter;
      val_middle = accumulated_load_trans / load_measures_counter;
      val_inner = accumulated_load_trans / load_measures_counter;
      accumulated_load_trans = 0;

    } else if (firstval == true) {
      firstval = false;
      accumulated_load_trans = 0;
      accumulated_load_outer = 0;
      accumulated_load_middle = 0;
      accumulated_load_inner = 0;

      val_trans = 0;
      val_inner = 0;
      val_middle = 0;
      val_outer = 0;
    }

    load_measures_counter = 0;
  }

  if (val_trans <= 180 && val_trans > 0) {
    driver_trans.VACTUAL(0);
    Serial.println("Stop, TRANS stall detected");
    val_trans = 0;
    runonce_trans = 0;
    firstval_trans = true;
    reverse_tip_stall = true;
  }

  if (val_outer <= 180 && val_outer > 0) {
    driver_outer.VACTUAL(0);
    Serial.println("Stop, OUTER stall detected");
    val_outer = 0;
    runonce_outer = 0;
    firstval_outer = true;
  }

  if (val_inner <= 50 && val_inner > 0 && k < 20) {
    driver_inner.VACTUAL(0);
    Serial.println("Stop, INNER 1 stall detected");
    val_inner = 0;
    runonce_inner = 0;
    firstval_inner = true;
    reverse_tip_stall = true;
  }

  if (val_inner <= 180 && val_inner > 0 && k > 22) {
    driver_inner.VACTUAL(0);
    Serial.println("Stop, INNER 2 stall detected");
    val_inner = 0;
    runonce_inner = 0;
    firstval_inner = true;
    reverse_tip_stall = true;
  }

  if (val_middle <= 180 && val_middle > 0) {
    driver_middle.VACTUAL(0);
    Serial.println("Stop, MIDDLE stall detected");
    val_middle = 0;
    runonce_middle = 0;
    firstval_middle = true;
  }
}

void reset_parameters() {
  // RESET ALL PROCESS FLAGS
  startRotate_CW = 0;
  startRotate_CCW = 0;
  startBend_for = 0;
  startBend_back = 0;
  startTip_close = 0;
  startTip_open = 0;
  startRotEnd_CCW = 0;
  startRotEnd_CW = 0;
  startCalibration = 0;
  startHoming = 0;
  startInstExchange = 0;
  startCouple = 0;

  initializeCalibration = 0;
  initializeHoming = 0;
  initializeInstExchange = 0;
  initializeCouple = 0;

  couple_process = 0;
  calibration_process = 0;
  instrexchange_process = 0;
  homing_process = 0;
  k = 0;

  val_outer = 0;
  cal_counter_outer = 0;
  timecal_outer = 0;
  firstval_outer = true;

  val_trans = 0;
  cal_counter_trans = 0;
  timecal_trans = 0;
  firstval_trans = true;

  runonce_outer = 0;
  runonce_middle = 0;
  runonce = 0;

  home_point_outer = 0;
  home_point_middle = 0;

  val_middle = 0;
  firstval_middle = true;

  hall_threshold_outer = false;
  hall_threshold_middle = false;

  home_outer_dir = true;
  home_middle_dir = true;

  outer_homed = false;
  middle_homed = false;
  trans_homed = false;

  outer_home_return = false;
  middle_home_return = false;

  outer_calibrated = false;
  trans_calibrated = false;

  flag = 1;
}